/**
 *  Mailerq.h
 *
 *  Public API for writing plugins
 *
 *  @author Michael van der Werve <michael.vanderwerve@mailerq.com>
 *  @copyright 2014 - 2019 Copernica B.V.
 */

/**
 *  Include guard.
 *  We don't use pragma once here for maximum portability
 */
#ifndef MQ_PLUGIN_EVENT_H
#define MQ_PLUGIN_EVENT_H

/**
 *  Plugins can be made in vanilla C
 */
#ifdef __cplusplus
extern "C" {
#endif

/**
 *  Macro for exporting public plugin functions
 */
#define MQ_EXPORT __attribute__ ((visibility("default")))

/**
 *  Standard type definitions
 */
#include <stdbool.h>
#include <stdint.h>
#include <stddef.h>

/**
 *  The different IO events we can be monitoring
 */
#define MQ_READ     1
#define MQ_WRITE    2

/**
 *  Log priorities
 */
#define MQ_CRIT     1
#define MQ_WARN     2
#define MQ_INFO     3

/**
 *  The current MailerQ API version
 *  @var    int
 */
#define MQ_API_VERSION  8

/**
 *  The MQ_Context structure provides access to the event loop, as well as the
 *  possibility to return control back to MailerQ after finishing work in
 *  asynchronous functions.
 */
struct MQ_Context;

/**
 *  The envelope and recipient data is encapsulated in an MQ_Address structure
 */
struct MQ_Address;

/**
 *  MQ_Message structure is used for plugins that operate on messages that
 *  have to be sent out.
 */
struct MQ_Message;

/**
 *  Internal MQ_Connection structure represents an incoming SMTP connection.
 */
struct MQ_Connection;

/**
 *  Structure representing an IO watcher.
 */
struct MQ_IOWatcher;

/**
 *  Timer context, returned from timer creation functions and required to
 *  cancel existing timer instances
 */
struct MQ_Timer;

/**
 *  JSON object structure. This is defined in the json-c library. To work with
 *  this structure, install and include the library in your plugin
 */
struct fjson_object;

/**
 *  Typedef all structures, so we need not type 'struct' everywhere.
 */
typedef struct MQ_Context       MQ_Context;
typedef struct MQ_Address       MQ_Address;
typedef struct MQ_Message       MQ_Message;
typedef struct MQ_Connection    MQ_Connection;
typedef struct MQ_IOWatcher     MQ_IOWatcher;
typedef struct MQ_Timer         MQ_Timer;

/**
 *  Log something in the MailerQ application-log. This will be prepended
 *  by the priority and the time in the final log entry.
 *  @param  priority
 *  @param  message
 */
void MQ_log(int priority, const char *message);

/**
 *  Retrieve the email address given an envelope structure
 *
 *  Be aware:   This function may return a nullptr in case no
 *              envelope address exists. This is most commonly
 *              seen in DSN messages.
 *
 *  @param  envelope    the envelope address
 *  @return const char *
 */
const char *MQ_email(MQ_Address *address);

/**
 *  Is this an address that has to be treated as a local address and published
 *  to the local queue, or to the regular inbox queue? Only meaningful for
 *  recipient addresses.
 *  @param  address     the address to retrieve the setting from
 *  @return bool
 */
bool MQ_local(MQ_Address *address);

/**
 *  Mark an address as being a local message that will be published to the local queue
 *  @param  address     the address to update the setting for
 *  @param  value       new valule
 */
void MQ_setLocal(MQ_Address *address, bool value);

/**
 *  Retrieve the message id
 *
 *  @param  message     the message to retrieve the message id from
 *  @return             the message id
 */
const char *MQ_messageID(MQ_Message *message);

/**
 *  Update the message id
 *
 *  @param  message     the message to update the message id for
 *  @param  messageid   the new message id for the message
 */
void MQ_setMessageID(MQ_Message *message, const char *messageid);

/**
 *  Retrieve the envelope address
 *
 *  @param  message the message to retrieve the envelope address from
 *  @return         the envelope address
 */
MQ_Address *MQ_envelope(MQ_Message *message);

/**
 *  Update the envelope address
 *
 *  @param  message     the message to set the envelope address in
 *  @param  envelope    actual envelope address
 */
void MQ_setEnvelope(MQ_Message *message, const char *envelope);

/**
 *  Retrieve the recipient
 *
 *  @param  message the message to retrieve the recipient from
 *  @return         the recipient address
 */
MQ_Address *MQ_recipient(MQ_Message *message);

/**
 *  Update the recipient address
 *
 *  @param  message     the message to set the recipient address in
 *  @param  recipient   actual recipient address
 */
void MQ_setRecipient(MQ_Message *message, const char *recipient);

/**
 *  Retrieve the MIME data
 *
 *  @param  message the message to retrieve the mime from
 *  @return         the MIME data
 */
const char *MQ_mime(MQ_Message *message);

/**
 *  Update the MIME data
 *
 *  @param  message     the message to set the envelope address in
 *  @param  mime        actual mime string
 */
void MQ_setMime(MQ_Message *message, const char *mime);

/**
 *  Retrieve JSON data from a message
 *
 *  @param  message the message to retrieve the JSON from
 *  @return json    pointer to the JSON object
 */
struct fjson_object *MQ_json(MQ_Message *message);

/**
 *  Retrieve file descriptor that is used for the SMTP connection
 *
 *  @param  connection  the connection in which the data was stored
 *  @return int
 */
int MQ_fileDescriptor(MQ_Connection *connection);

/**
 *  Get whether the connection is secure or not
 *
 *  @param  connection  the connection that we want to know the status of
 *  @return bool
 */
bool MQ_secure(MQ_Connection *connection);

/**
 *  Get whether the connection is considered authenticated
 *
 *  @param  connection  the connection that we want to know the status of
 *  @return bool
 */
bool MQ_authenticated(MQ_Connection *connection);

/**
 *  Set whether the connection should be considered authenticated
 *
 *  @param  connection  the connection that authentication should be set for
 *  @param  bool        should the credentials be considered valid
 */
void MQ_setAuthenticated(MQ_Connection *connection, bool authenticated);

/**
 *  Send data over the SMTP connection
 *  @param  connection  the connection to send data over
 *  @param  buffer      buffer to send
 *  @param  size        size of the buffer
 *  @return size_t
 */
size_t MQ_send(MQ_Connection *connection, const char *buffer, size_t size);

/**
 *  Retrieve userdata stored using MQ_SetSmtpData
 *
 *  @param  connection  the connection in which the data was stored
 *  @return void*
 */
void *MQ_connectionData(MQ_Connection *connection);

/**
 *  Store userdata in the connection
 *
 *  @param  connection  the connection in which to store the data
 *  @param  void*       the data to store
 */
void MQ_setConnectionData(MQ_Connection *connection, void *data);

/**
 *  Retrieve the underlying libev event loop from the given context
 *
 *  @param  context the context to retrieve the loop from
 *  return  struct libev
 */
struct ev_loop *MQ_ev(MQ_Context *context);

/**
 *  Monitor a file descriptor for activity
 *
 *  @param  context     the context to work with
 *  @param  fd          the file descriptor to watch
 *  @param  events      the events to watch, binary or'ed MQ_READ and/or MQ_WRITE
 *  @param  callback    callback that will be executed when one of the events
 *                      occurs, the given data pointer is also provided to
 *                      the callback function
 *  @param  data        customer user data to supply to the callback
 *  @return MQ_IOWatcher
 */
MQ_IOWatcher *MQ_ioWatch(MQ_Context *context, int fd, int events, void(*callback)(MQ_IOWatcher *watcher, int fd, int events, void *data), void *data);

/**
 *  Stop watching a file descriptor for activity
 *
 *  @param  watcher     the watcher object
 */
void MQ_ioUnwatch(MQ_IOWatcher *watcher);

/**
 *  Start a timer.
 *
 *  @param  context     the context to work with
 *  @param  timeout     how long to wait before the timer expires
 *  @param  callback    the callback to invoke when the timer expires
 *  @param  data        customer user data to supply to the callback
 */
MQ_Timer *MQ_startTimer(MQ_Context *context, float timeout, void(*callback)(void *data), void *data);

/**
 *  Stop a timer prematurely
 *
 *  Note that the callback will not be invoked if the timer is stopped this way.
 *  If the timer has already expired, this function should *NOT* be called.
 *
 *  @param  timer   the timer to stop
 */
void MQ_stopTimer(struct MQ_Timer *timer);

/**
 *  Reset a timer to a different timeout
 *
 *  @param  timer   the timer to reset
 *  @param  timeout the new timeout
 */
void MQ_resetTimer(MQ_Timer *timer, float timeout);

/**
 *  Finish plugin processing
 *
 *  @param  connection  the connection that is now finished
 */
void MQ_complete(MQ_Connection *connection);

/**
 *  Finish plugin processing
 *
 *  @param  connection  the connection that may move on to the next plugin
 */
void MQ_continue(MQ_Connection *connection);

/**
 *  Retry plugin processing
 *
 *  @param  connection  the connection that should be retried
 */
void MQ_retry(MQ_Connection *connection);

/**
 *  Retrieve userdata stored using MQ_SetContextData
 *
 *  @param  context the context to retrieve the data from
 *  @return void*
 */
void *MQ_contextData(MQ_Context *context);

/**
 *  Store userdata in the context
 *
 *  @param  context the context to store the data in
 *  @param  void*   the data to store
 */
void MQ_setContextData(MQ_Context *context, void *data);

/**
 *  End of extern C declaration
 */
#ifdef __cplusplus
}
#endif

/**
 *  End of include guard
 */
#endif /* MQ_PLUGIN_EVENT_H */
